    ;; YaPPP 0.7 - Yet Another PPP.  Based on Thomas Neuhaus' PPP.
    ;; Copyright (C) 2013 Emanuel Wittersheim

    ;; This program is free software: you can redistribute it and/or modify
    ;; it under the terms of the GNU General Public License as published by
    ;; the Free Software Foundation, either version 3 of the License, or
    ;; (at your option) any later version.

    ;; This program is distributed in the hope that it will be useful,
    ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
    ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    ;; GNU General Public License for more details.

    ;; You should have received a copy of the GNU General Public License
    ;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; object-framework.scm
;;
;; definiert ein einfaches Objekt Framework und 
;; einige, mglicherweise ntzliche Klassen

;; Framework:
;;
; delegate reicht die Nachricht an das Basisobjekt weiter:

(define (delegate obj msg)
  (apply obj msg))

; base-object versteht nur type:

(define base-object
  (lambda msg
    (case (car msg)
      ((type) "base-object")
      (else invalid-message-indicator))))

; send: Sende an ein Objekt eine Machricht

(define send
  (lambda args
    (let* ((object (car args)) 
	   (message (cdr args))
	   (try (apply object message)))
      (if (eq? invalid-message-indicator try)
	  (error (string-append (symbol->string (car message)) ":"
				"Bad Message send to a "
				(object 'type)))
	  try))))

(define invalid-message-indicator "unknown")

;; Einfache Klassen:
;;
;; a simple Container

(define (box-maker init-value)
  (let ((contents init-value))
    (lambda msg
      (case (car msg)
	((type) "box")
	((show) contents)
	((update!) (set! contents (cadr msg)))
	((swap!) (let ((result contents))
		   (set! contents (cadr msg))
		   result))
	((reset!) (set! contents init-value))
	(else
	 (delegate base-object msg))))))


;; Ein "counter" ist ein Object, dass seinen Wert nach einer festen Regel
;; aendert, z.b. indem es 1 auf seinen alten wert addiert.
;; Ein Counter hat auch einen initialen Wert.
;; Counter reagieren auf update! show reset! type
;; counter is A Box

(define (counter-maker init-value change-proc)
  (let ((total (box-maker init-value)))
    (lambda msg
      (case (car msg)
	((type) "counter")
	((update!) (let ((result (change-proc (send total 'show))))
		    (send total 'update! result)))
	(else
	 (delegate total msg))))))

;; Bei einem "accumulator" bekommt die update! prozedur noch ein weiteres 
;; Argument
;; Accumulator is A Box (is A Counter ?)
(define (accumulator-maker init-value change-proc)
  (let ((total (box-maker init-value)))
    (lambda msg
      (case (car msg)
	((type) "accumulator")
	((update!) (let ((result (change-proc (send total 'show) (cadr msg))))
		    (send total 'update! result)))
	(else
	 (delegate total msg))))))


;; bei einem "gauge" (Messgeraet, Messfuehler...) gibt es eine Update prozedur
;; up! und eine update-prozedur !down

(define (gauge-maker init-value change-up change-down)
  (let ((total (box-maker init-value)))
    (lambda msg
      (case (car msg)
	((type) "gauge")
	((up!) (let ((result (change-up (send total 'show))))
		    (send total 'update! result)))
	((down!) (let ((result (change-down (send total 'show))))
		    (send total 'update! result)))
	((update!) (error "No Update! with Gauges"))
	(else
	 (delegate total msg))))))

;; schreibe acc-max, ein accumulator,  mit initialwert 0, der bei jedem update
;; den aktuellen wert mit dem neuen vergleicht und nur den hoeheren speichert

(define (acc-max-maker init-value change-proc)
  (let ((total (accumulator-maker init-value change-proc)))
    (lambda msg
      (case (car msg)
	((type) "acc-max")
	((update!) (let* ((oldval (send total 'show))
			  (result (change-proc oldval (cadr msg))))
		    (if (< oldval result)
			(send total 'update! (cadr msg)))))
	(else
	 (delegate total msg))))))

;; schreibe einen restricted counter, der zu seiner update-prozedur auch
;; noch ein praedikat erhält, und dafuer sorgt, dass der gespeicherte wert immer
;; das Praedikat erfuellt. Tut ein wert das nicht, erfolgt ein reset!

(define (restricted-counter-maker init-value change-proc pred)
  (let ((total (counter-maker init-value change-proc)))
    (lambda msg
      (case (car msg)
	((type) "restricted-counter")
	((update!) (begin
		     (send total 'update!)
		     (if (not (pred (send total 'show)))
			 (send total 'reset!))))
	(else
	 (delegate total msg))))))


(newline)
(display "YaPPP 0.7 Copyright (C) 2013 Emanuel Wittersheim
	This program comes with ABSOLUTELY NO WARRANTY. For details see file COPYING. 
	This is free software, and you are welcome to redistribute it
	under certain conditions.")
(newline)
